#!/usr/bin/perl
#
# Copyright(c) 2006-2019 ThreatSTOP, Inc. All Rights Reserved
#
# NOTICE: All information contained herein is, and remains the property
# of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to ThreatSTOP, Inc.
# and its suppliers and certain aspects thereof are protected by United
# States Patent No. 8,533,822 and United States Patent No. 8,869,237,
# and are protected by trade secret or copyright law.
#
# US Government Entities: The ThreatSTOP service, software and
# documentation, as applicable, are "commercial computer software" and
# "commercial computer software documentation" developed exclusively at
# private expense by Threatstop, Inc. ("Threatstop"). Pursuant to FAR
# 12.212 or DFARS 227 7202 and their successors, as applicable, use,
# reproduction and disclosure of the software and documentation is
# governed by the terms of Threatstop's commercial agreement.
#
# ThreatSTOP main function module
################################################################################

package ThreatStop;
use Exporter;

use strict;
use warnings;
use Scalar::Util qw(openhandle);
use Cwd qw(abs_path);
use Fcntl qw(:flock);
use File::Basename;
use File::Copy; # provides move, copy
use Fcntl ':mode'; # used for get_file_stat()
use FindBin;
# use Switch;
use Data::Dumper;
use MIME::Base64 qw(encode_base64 decode_base64);
# FindBin->again;
use POSIX qw(strftime setuid);

my $USAGE; # must define here, but overridden in tsadmin or $0 script

###############################################################
################## MODULE DEFENITIONS PART ####################
###############################################################
our $VERSION = 1.00;
our @ISA     = qw(Exporter);
our @EXPORT =
  qw(seteuid get_f_telemetry get_f_telemetry_value get_f_telemetry_elapsed clear_f_telemetry set_f_telemetry check_for_app get_file_mtime find_app_conf find_app_confd app_is_running get_app_version get_app_restart detect_app_info install_via_package_manager get_package_manager_install_cmd check_for_module install_via_cpan install_config install_app_config uninstall_app_config is_ctp_install abortIfNotUser get_or_create_dir add_to_file remove_from_file write_conf_key write_conf_keys detect_os get_UID rewriteConfKeys confirm_prompt input_prompt print_banner print_colorized getCurrentTime logInit printMsg logMsg setLock setUnLock ipv4_cidr2msk checkDNSaccess calculateElapsedTime loadConfig intHandler osPlatformDetecter loadConfFileData setWorkingDir setAlarm osDestributionDetector appendUploadLogEntry purgeOldUploadsLogs $SUDO $TSDIR $TSLOGDIR $DNS_LIST $DEBUG_LEVEL $LOGFILE $LOG_LEVEL $dns_server $port $LOCKFILE backup_file restore_file generate_sha1_checksum);

# our $TSDIR = dirname abs_path($0); # gets absolute path to this file's dir
our $TSDIR = '/opt/threatstop'; # gets absolute path to this file's dir
our $TSLOGDIR = '/var/log/threatstop'; # gets absolute path to log dir

our $CFG_DATA;
our $CONFIGFILE = $TSDIR . "/threatstop.conf"
  if ( !defined($CONFIGFILE) );
our $LOGFILE = $TSDIR . "/tsoutput.log"
  if ( !defined $LOGFILE );    #Define default log
our $UPLOADSLOGFILE = $TSLOGDIR . "/tsuploads";
our $LOCKFILE       = $TSDIR . '/ts-setup.lck' if ( !defined($LOCKFILE) );
our $LOGINIT        = 0;
our $DEBUG_LEVEL    = 4
  if ( !defined $DEBUG_LEVEL );    #Define verbose level 5-nothing
if (@ARGV) {
    $DEBUG_LEVEL = 1 if ( grep(/^SILENT$/, @ARGV) );
    $DEBUG_LEVEL = 6 if ( grep(/^DEBUG$/, @ARGV) );
}
our $LOG_LEVEL = 6 if ( !defined $LOG_LEVEL );   #Define verbose level 5-nothing
our ( $FD, $FDL, $LFD );
our $START_EPOCH = time();
our $END_EPOCH;

our $DNS_LIST = "192.124.129.42";
our $dns_server;                                 # "192.124.129.42";
our $DNS_PROTO     = '+tcp';                     # Protocol to use - TCP
our $DNS_PROTO_ALT = '+notcp';                   # Protocol to use - UDP
our $DNS_PORT_BASE = 53;                         # Port to use (default 53)
our $DNS_PORT_ALT  = 5353;                       # Port to use (alternate 5353)
our $DNS_DOMAIN_BASE = 'dns-servers.threatstop.local';  # Server name to resolve
our $DNS_DOMAIN_ALT =
  'dns-servers.threatstop.com';    # Alternative server name to resolve
our $DNS_SERVERS_ALT = "8.8.8.8 8.8.4.4 208.67.220.220 208.67.222.222"
  ;                                # Alternate servers ip's to quiery
our $DNS_SERVERS_LCL = '127.0.0.1';    # Local DNS serves will be stored here
our $DNS_LIST_NEW;
our $DNS_PORT = 53;    # Actually we read this param from conf file
our $curr_date = strftime "%m%Y", localtime;

our $SUDO = `which sudo`;
if ($? != 0) {
    $SUDO = ''
} else {
    chomp($SUDO);
}

# Setup signal handler for possible user interupts
$SIG{'INT'}  = 'intHandler';
$SIG{'ABRT'} = 'intHandler';
$SIG{'QUIT'} = 'intHandler';
$SIG{'STOP'} = 'intHandler';

###############################################################
################# FUNCTIONS DEFENITIONS PART ##################
###############################################################

############################################
########## PRIVATE FUNCTIONS ###############
############################################

################################################################

##################################################
# intHandler : Used to trap exit signals and per #
#              form lockfile cleanup             #
##################################################
sub intHandler() {
    if ( defined $FDL ) {
        close($FDL);
        unlink($LOCKFILE);
    }

    print "\n";
    exit(0);
}

sub get_log_hhmmss {
    return strftime("%H:%M:%S - ", localtime);
}

##################################################
# loadConfig : Loads additional parameters from  #
#              conf file                         #
##################################################
sub loadConfig() {

    &printMsg( "INFO", "Loading configuration file ..." );

    eval {
        open( FD, "<$CONFIGFILE" )
          or die
"Failed to open configuration file [$CONFIGFILE].  The following error occurred [$!]";
    };

    if ($@) {
        &printMsg( "ERROR", "$@" );
        return 1;
    }

    while (<FD>) {
        chomp;       # no newline
        s/#.*//;     # no comments
        s/^\s+//;    # no leading white
        s/\s+$//;    # no trailing white
        s/'//g;      # no ticks
        next unless length;    # anything left?
        my ( $var, $value ) = split( /\s*=\s*/, $_, 2 );
        $CFG_DATA->{$var} = $value;
    }

    close FD;
    mpw( $CONFIGFILE, $CFG_DATA );
    return 0;
}

#mangle/unmangle config must be called right after loading in config values.
sub mpw {
    my $conf    = shift;
    my $confref = shift;
    foreach my $i (qw(password enable_pw)) {
        if ( defined( $confref->{$i} ) ) {
            my $tmp = $confref->{$i};
            $tmp =~ s/\"//g;
            if ( $tmp =~ /^\$\$.*\$\$$/ ) {
                $tmp =~ s/\$//g;
                $tmp =~ s/!/=/g;
                $tmp = decode_base64($tmp);

                #               print "$i encoded\n";
                #               print "$i: $tmp\n";
                $confref->{$i} = $tmp;
            }
            else {
                #               print "$i plaintext\n";
                #               print "$i: $confref->{$i}\n";
                my $tmp = $confref->{$i};
                $tmp =~ s/\"//g;
                $tmp = encode_base64($tmp);
                chomp($tmp);
                $tmp =~ s/=/!/g;
                $tmp = '"' . '$$' . $tmp . '$$' . '"';
                rewriteConfKey( $conf, $i, $tmp );
            }

        }
    }
}

=head3 write_conf_key

 Write the given key/value pair in the specified config file. Adds it if it is not found.
 NOTE: adds a commented description from usage hash if exists just before key:value pair

args:
    string - file path to config
    hash - key:value pair
    string - add_if_missing 0 - don't add if missing, any positive means true, add if missing. *defaults to true
returns:
    1 - fail
    0 - success
=cut
sub write_conf_key {
    my $conf  = shift;
    my $kvpair   = shift;
    my $add_if_missing = shift || 1;
    # exit early
    if (scalar (keys %{$kvpair}) != 1) {
        &printMsg('ERROR', "expected 1 key value pair to write");
        return 1;
    }
    if (! -f $conf or ! -w $conf) {
        &printMsg('ERROR', "expected config exists and writable");
        return 1;
    }

    my $key   = (keys %{$kvpair})[0]; # get first key is key:value hash
    my $value = $kvpair->{$key};
    my $in;
    my $out;
    my @data;
    eval {
        open( $in, '<', $conf );
        @data = <$in>;
        close($in);
        open( $out, '>', $conf );
    };
    if ($@) {
        &printMsg( "ERROR", $@ );
        return 1;
    }
    my $wrote_line = 0;
    foreach my $line (@data) {
        if ( $line =~ /^$key=/ ) {
            $wrote_line++;
            print $out qq|$key="$value"\n|;
            # &printMsg( "DEBUG", "Rewrote $key = $value in $conf" );
        }
        else {
            print $out $line;
        }
    }
    unless ($wrote_line) {
        # if we have a description of the key, add that commented out via '#'
        if (defined $USAGE and exists $USAGE->{$key}) {
            print $out qq|# $USAGE->{$key}"\n|;
            # &printMsg( "DEBUG", "Added comment: # $USAGE->{$key} " );
        }
        print $out qq|$key="$value"\n|;
        &printMsg( "DEBUG", "Wrote: $key = $value in $conf" );
    }
    close($out);
    return 0;

}


=head3 write_conf_keys

 Write the given key/value pair hash in the specified config file. Adds it if it is not found.
 NOTE: adds a commented description from usage hash if exists just before key:value pair
 Comments are defined as lines starting with a '#' or ';'

args:
    string - file path to source config
    string - file path to target config (*can be the same)
    hash - key:value pair
    string - add_if_missing 0 - don't add if missing, any positive means true, add if missing. *defaults to true
returns:
    1 - fail
    0 - success
=cut
sub write_conf_keys {
    my $source_conf  = shift;
    my $target_conf  = shift;
    my $kvpairs   = shift;
    my $usage   = shift; # optional usage hash to provide comments just before newly added conf option

    # &printMsg('DEBUG', "write_conf_keys called: $source_conf, $target_conf");
    # &printMsg('DEBUG', "write_conf_keys kvpairs: " . Dumper $kvpairs);
    # exit early
    if (! -f $target_conf) {
        &printMsg('INFO', "Creating blank configuration file: $target_conf");
        eval { qx(touch $target_conf); };
        if ($@) {
            &printMsg('INFO', "Creating blank configuration file: $target_conf");
            &printMsg('ERROR', "Error in creating conf '$target_conf' Err:" . $@);
            return 1;
        }
    }
    if (ref($kvpairs) ne "HASH" and ref($kvpairs) ne "HASHREF") {
        &printMsg('ERROR', "expected key value pair hash ref");
        return 1;
    }
    if (! -f $source_conf or ! -w $target_conf) {
        &printMsg('ERROR', "expected source config exists and target conf writable by effective UID:GID");
        return 1;
    }


    my $add_if_missing = shift || 1;
    my $in;
    my $out;
    my @data;
    eval {
        open( $in, '<', $source_conf );
        @data = <$in>;
        close($in);
        open( $out, '>', $target_conf );
    };
    if ($@) {
        &printMsg( "ERROR", $@ );
        return 1;
    }
    my $updated = {};
    foreach my $line (@data) {
        # &printMsg('DEBUG', "LINE: $line");
        $line =~ /^([a-z]\w+)=\s?[("]?(.*?)[")]?\s+$/;
        my $conf_key = $1;
        my $conf_value = $2;
        unless ($conf_key) {
            # we have comment or whitespace.. print it out
            print $out $line;
            next;
        }
        if ( exists $kvpairs->{$conf_key} ) {
            $updated->{$conf_key} = 1;
            my $value = $kvpairs->{$conf_key};
            if (ref($value) eq 'HASH' or ref($value) eq 'HASHREF') {
                # &printMsg( "DEBUG", "Refusing to write $conf_key. Value is a hash." . Dumper $value );
                &printMsg( "WARN", "Refusing to write $conf_key. Value is a hash.");
                next;
            }
            print $out qq|$conf_key="$value"\n|;
            # &printMsg( "DEBUG", "Rewrote $conf_key, '$conf_value' -> '$value' in $target_conf" );
            next;
        }
        else {
            print $out $line;
        }
    }
    # loop again and if not seen in update hash, add key:value
    for my $add_key (sort (keys %{$kvpairs})) {
        unless (exists $updated->{$add_key}) {
            # if we have a description of the key, add that commented out via '#'
            if (defined $usage and exists $usage->{$add_key}) {
                print $out qq|# $usage->{$add_key}\n|;
                # &printMsg( "DEBUG", "Added comment: # $usage->{$add_key} " );
            }
            print $out qq|$add_key="$kvpairs->{$add_key}"\n\n|;
            # &printMsg( "DEBUG", "Added $add_key = $kvpairs->{$add_key} in $target_conf" );
        }
    }
    close($out);
    &printMsg('OK', "wrote $target_conf");
    return 0;

}

#
# rewriteConfKeys()
#
# Args :
#   src -
#   target -
#   keys -
#
# Returns :
#   0 - success
#   1 - error
#
# public function for generating/regenerating a config file based
# on a key/value hash and a template file.
# rewriteConfKeys(sourceConfile,targetConfile,\%keys)
# read in the source config file and based on the keys hashref
# create or re-write the target config file with new key=value pairs
# for each key inside the keys hashref. key=value pairs that are not
# in the hashref or are already the same value as keys in the hashref
# remain untouched.
#
# to generate from a template use the template as the source and the
# new config as the target.  To re-configure an existing file simply
# use the same file for both source and target.
#
sub rewriteConfKeys {
    my $src    = shift;
    my $target = shift;
    my $keys   = shift;
    my $in;
    my $out;
    my @data;

    eval {
        open( $in, '<', $src );
        @data = <$in>;
        close($in);
        open( $out, '>', $target );
    };
    if ($@) {
        printMsg( "ERROR", $@ );
        return 1;
    }
    foreach my $line (@data) {
        my $found = 0;
        foreach my $key ( keys %$keys ) {

            my $value = $keys->{$key};
            if ( $line =~ /^$key=/ ) {
                $line =~ /^([a-z]\w+)=[("]?(.*?)[")]?\s+$/;
                my $tmp = $2;
                $tmp =~ s/\"//g;
                last if $tmp eq $value;
                print $out "$key=\"$value\"\n";
                printMsg( "DEBUG", "rewrote $key = $value in $target" );
                $found = 1;
            }
        }
        print $out $line if !$found;

    }
    close($out);
    return 0;

}

#####################################################################
sub readLocalBoxNameserversLst() {

    my $servers_lst = shift;
    my $cmd =
      "cat /etc/resolv.conf | grep nameserver | /usr/bin/awk '{print \$2}'";
    my $rc = qx($cmd);

    &printMsg( "INFO", "Reading local box configuration." );
    $rc =~ tr/\n/ /;
    $$servers_lst = $rc;

    return 0;
}

############################################
########### PUBLIC FUNCTIONS ###############
############################################

##################################################
# getCurrentTime : Retruns curretn time stamp    #
##################################################
sub getCurrentTime() {

    my $tr = shift;
    my @td = localtime(time);
    my $ts = sprintf(
        "%02d/%02d/%04d %02d:%02d:%02d",
        $td[4] + 1,
        $td[3],
        $td[5] + 1900,
        $td[2], $td[1], $td[0]
    );

    $$tr = $ts;

    return 0;
}

##################################################
#logInit : initialize logging mechanism          #
##################################################
sub logInit {

    my $logfile = shift;

    #If logfile supplied use it instead of default
    $LOGFILE = $logfile if ( defined $logfile );
    $LOGINIT = 1;

    #sometimes the script gets run as root by hand and we
    #wind up with an unwritable log file this check catches this case
    #and attempts to unlink the file before we open for write.
    if ( -f $LOGFILE ) {
        if ( !-w $LOGFILE ) {
            &printMsg( "WARN",
                "Couldn't open $LOGFILE for writing.  Attempting to unlink." );
            $LOGINIT = 0;
            if ( unlink($LOGFILE) ) {
                &printMsg( "WARN", "Successfully unlinked $LOGFILE" );
            }
        }
    }
    eval { open( LFD, '>' . $LOGFILE ) or die $!; };

    if ($@) {
        &printMsg( "WARN",
            "Can not initialize logging ... Proceeding without ..." );
        $LOGINIT = 0;
        return 0;
    }

    return 0;
}

##################################################
#logMsg : prints info into log file              #
##################################################
sub logMsg {

    my $log_type = shift;
    my $log_body = join( ' ', @_ );
    my $log_out;

    if ($log_type eq 'ERROR') { $log_out = "[ERROR] : $log_body"; }
    elsif ($log_type eq 'ALERT') {
        $log_out = "[ALERT] : $log_body"
          if ( $LOG_LEVEL >= 1 );
    }
    elsif ($log_type eq 'WARN') { $log_out = "[WARN ] : $log_body" if ( $LOG_LEVEL >= 2 ); }
    elsif ($log_type eq 'INFO') { $log_out = "[INFO ] : $log_body" if ( $LOG_LEVEL >= 3 ); }
    elsif ($log_type eq 'DISPL') { $log_out = $log_body if ( $LOG_LEVEL >= 4 ); }
    elsif ($log_type eq 'NOTE') { $log_out = "[NOTE ] : $log_body" if ( $LOG_LEVEL >= 5 ); }
    elsif ($log_type eq 'DEBUG') {
        $log_out = "\t[DEBUG] : $log_body" if ( $LOG_LEVEL >= 6 );
    }

    if ( defined $log_out ) {
        print LFD $log_out . "\n" if ( $LOGINIT == 1 );
    }

    return 0;
}

##################################################
# printMsg : Prints/logs different level output  #
##################################################
sub printMsg {

    my $msg_type = shift;
    my $msg_body = join( ' ', @_ );
    my $msg_out;


    if ($msg_type eq 'DEBUG') {
        # $msg_out = "\t[DEBUG] : $msg_body" if ( $DEBUG_LEVEL >= 6 );
        $msg_out = get_log_hhmmss() . sprintf('%-62s%s', $msg_body, "[DEBUG]") if ( $DEBUG_LEVEL >= 6 );
    }
    elsif ($msg_type eq 'ALERT') {
        # $msg_out = "[ALERT] : $msg_body" if ( $DEBUG_LEVEL >= 1 );
        $msg_out = get_log_hhmmss() . sprintf('%-62s%s', $msg_body, "[ALERT]\n") if ($DEBUG_LEVEL >= 1);
    }
    elsif ($msg_type eq 'WARN') {
        # $msg_out = "[WARN ] : $msg_body" if ( $DEBUG_LEVEL >= 2 );
        $msg_out = get_log_hhmmss() . sprintf('%-62s%s', $msg_body, "[WARN]\n") if ($DEBUG_LEVEL >= 2);
    }
    elsif ($msg_type eq 'INFO') {
        # $msg_out = "[INFO ] : $msg_body" if ( $DEBUG_LEVEL >= 3 );
        $msg_out = get_log_hhmmss() . sprintf('%-62s%s', $msg_body, "") if ($DEBUG_LEVEL >= 3);
    }
    elsif ($msg_type eq 'DISPL') {
        $msg_out = $msg_body if ( $DEBUG_LEVEL >= 4 );
    }
    elsif ($msg_type eq 'NOTE') {
        # $msg_out = "[NOTE ] : $msg_body" if ( $DEBUG_LEVEL >= 5 );
        $msg_out = get_log_hhmmss() . sprintf('%-62s%s', $msg_body, "[NOTE]");
    }
    elsif ($msg_type eq 'ERROR') {
        # $msg_out = "[ERROR] : $msg_body";
        $msg_out = get_log_hhmmss() . sprintf('%-62s%s', $msg_body, "[ERROR]\n");
    }
    elsif ($msg_type eq 'OK') {
        $msg_out = get_log_hhmmss() . sprintf('%-62s%s', $msg_body, "[OK]") if ( $DEBUG_LEVEL >= 5 );
    }
    elsif ($msg_type eq 'FAIL') {
        $msg_out = get_log_hhmmss() . sprintf('%-62s%s', $msg_body, "[FAIL]") if ( $DEBUG_LEVEL >= 5 );
    }
    elsif ($msg_type eq 'UP') {
        $msg_out = get_log_hhmmss() . sprintf('%-62s%s', $msg_body, "[UP]") if ( $DEBUG_LEVEL >= 5 );
    }
    elsif ($msg_type eq 'DOWN') {
        $msg_out = get_log_hhmmss() . sprintf('%-62s%s', $msg_body, "[DOWN]") if ( $DEBUG_LEVEL >= 5 );
    }
    elsif ($msg_type eq 'START') {
        $msg_out = get_log_hhmmss() . sprintf('%-62s%s', $msg_body, "[START]") if ( $DEBUG_LEVEL >= 5 );
    }
    elsif ($msg_type eq 'DOWN') {
        $msg_out = get_log_hhmmss() . sprintf('%-62s%s', $msg_body, "[DOWN]") if ( $DEBUG_LEVEL >= 5 );
    }
    elsif ($msg_type eq 'CHECK') {
        $msg_out = get_log_hhmmss() . sprintf('%-62s%s', $msg_body, "[CHECK]") if ( $DEBUG_LEVEL >= 5 );
    }
    else {
        $msg_out = "[SYS_ERROR] : Unrecognized message type [$msg_type]";
    }

    &logMsg( "$msg_type", "$msg_body" ) if ( $LOGINIT == 1 );

    if ( defined $msg_out ) {
        print $msg_out . "\n";
    }

    return 0;
}

sub is_valid_ip {
    my $ip = shift;
    if ($ip =~ m/^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/) {
        return 1;
    }
    return 0;
}
=head3 getMyIP

 Return IP of the non-localhost interface

=cut

sub getMyIP {

    my $myip  = '0.0.0.0';
    my $ipcmd = undef;
    for my $ip ( '/bin/ip', '/sbin/ip', '/usr/sbin/ip' ) {
        if ( -e $ip ) {
            $ipcmd = $ip;
            last;
        }
    }
    if ( !defined $ipcmd ) {
        printMsg( "DEBUG", "Unable to find 'ip' application" );
    } else {
        if ( open( SHOWIP, "-|", "$ipcmd addr show" ) ) {
            my $if = undef;
            while (<SHOWIP>) {

    # looking for start of new interface line, like
    #  2: ens160: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc mq state UP group default qlen 1000

                if (/^\d+: ([^:]+): /) {

                    $if = $1;
                    if ( $if eq 'lo' ) {

                        $if = undef;
                    }
                }

                # will wait until we're "in" a non-localhost interface section
                if ( !defined $if ) {
                    next;
                }

            # looking for ip address line, like
            #     inet 172.21.88.59/24 brd 172.21.88.255 scope global dynamic enp0s3
                if (/^\s+inet ([0-9.]+)\/\d+ /) {

                    # will use first non localhost interface we find
                    $myip = $1;
                    printMsg( "DEBUG", "Determined local IP address to be $myip" );
                    last;
                }
            }
            close SHOWIP;
        }
    }


    # fetching first interfaces IP (pfsense support)
    # em0: flags=8843<UP,BROADCAST,RUNNING,SIMPLEX,MULTICAST> metric 0 mtu 1500
    # options=9b<RXCSUM,TXCSUM,VLAN_MTU,VLAN_HWTAGGING,VLAN_HWCSUM>
    # ether 08:00:27:b3:f9:fb
    # hwaddr 08:00:27:b3:f9:fb
    # inet6 fe80::a00:27ff:feb3:f9fb%em0 prefixlen 64 scopeid 0x1
    # inet 10.0.2.15 netmask 0xffffff00 broadcast 10.0.2.255
    chomp(my $ifconfig = `which ifconfig`);
    &printMsg('DEBUG', "ifconfig $ifconfig");
    if ($ifconfig) {
        my @possible_interfaces = qw(em0 em1 em2 eth0 eth1 eth2 br0);
        for my $int (@possible_interfaces) {
            &printMsg('DEBUG', "Attempting to fetch $int IP address");
            if (chomp($myip = `$ifconfig $int|grep 'inet '|awk '{print \$2}'`)) {
                &printMsg('DEBUG', "Interface [$int] returned '$myip'");
                last;
            }
        }
    }

    if (!is_valid_ip($myip)) {
        printMsg( "ERROR", "Unable to examine system IP addresses" );
    }


    return $myip;
}


##################################################
# setLock : Locks running instance               #
##################################################
sub setLock {

    &printMsg( "DEBUG", "Locking current execution instance." );

    open( $FDL, ">>$LOCKFILE" )
      or die
"Failed to touch lockfile [$LOCKFILE].  The following error occurred [$!]";
    unless ( flock( $FDL, LOCK_EX | LOCK_NB ) ) {
        &printMsg( "ALERT", "$0 is already running... Exiting ... lockfile: $LOCKFILE" );
        exit(1);
    }

    return 0;
}

##################################################
# setUnLock : Unlocks running instance           #
##################################################
sub setUnLock {

    &printMsg( "DEBUG", "Unlocking current execution instance." );

    if ( openhandle($FDL) ) {
        flock( $FDL, LOCK_UN );
        close($FDL);
        unlink($LOCKFILE);
    }

    return 0;
}

####################################################
# setWorkingDir : set running directory to be local#
#                 used for non local path run      #
####################################################
sub setWorkingDir() {

    my $mypath = abs_path($0);
    my $mydir  = dirname($mypath);

    &printMsg( "DEBUG", "Switching to [$mydir] working directory." );

    eval { chdir $mydir or die "Failed to change working directory."; };

    if ($@) {
        &printMsg( "ERROR", "$@" );
        return 1;
    }
    return 0;
}

#This subroutine works very similar to alarm except it is itself safe to use
#with code that itself calls alarm or sleep. it does this by setting up a signal handler for SIGUSR1
#and then forking a process that waits the specified amount of time before giving us the SIGUSR1 signal.
#Since forked processes do no inherit the alarm(2) timer it is safe to use this timer without resetting
#any previously set timers in the main process or vice-versa. We also restore anything previously saved into $SIG{USR1} if anything
#when the timer is reset with a 0 interval. Optionally if a subroutine reference is passed as the second argument
#it will be used as the SIGUSR1 handler instead of the default.
{
    my $child = 0;
    my $oldusr1;

    sub setAlarm {
        my $interval   = shift;
        my $handler    = shift;
        my $defhandler = sub { die "ThreatStop::setAlarm() Timer Expired."; };

        #Set up signal handler
        if ( !$handler ) {
            $handler = $defhandler;
        }

#Save existing USR1 handler if it exists.
#but only do this once so we don't lose the original with multiple calls to setAlarm.
        if ( ( !$oldusr1 ) and $SIG{USR1} ) {
            if ( $SIG{USR1} != $defhandler )
            {    #Make sure we don't "restore" our own internal handler later.
                &printMsg( "DEBUG",
"ThreatStop::setAlarm() saving existing USR1 handler for later."
                );
                $oldusr1 = $SIG{USR1};
            }
        }
        $SIG{USR1} = $handler;

        if ( !$interval ) {    #handle unsetting the alarm and signal handler.
            &printMsg( "DEBUG", "ThreatStop::setAlarm() unsetting alarm." );
            if ($child) {
                &printMsg( "DEBUG",
                    "ThreatStop::setAlarm() killing child $child" );
                kill( "TERM", $child );
                $child = 0;
            }
            if ($oldusr1) {
                &printMsg( "DEBUG",
                    "ThreatStop::setAlarm() restoring saved signal handler." );
                $SIG{USR1} = $oldusr1;
                $oldusr1 = undef;
            }
            else { $SIG{USR1} = undef; }
            return 0;
        }
        else {    #Fork the simple process to signal us at a later time.
            if ($child) {
                &printMsg( "DEBUG",
                    "ThreatStop::setAlarm() killing child $child" );
                kill( "TERM", $child );
            }
            my $pid = fork();
            if ($pid) {
                &printMsg( "DEBUG",
                    "ThreatStop::setAlarm() forked child $pid" );
                $child = $pid;
                return 0;
            }
            else {
                &printMsg( "DEBUG",
"ThreatStop::setAlarm() child setting alarm to $interval seconds."
                );
                close($FDL)
                  if openhandle($FDL);   #alarm processes should not hold locks.
                sleep $interval;
                kill( "USR1", getppid() );
                exit(0);
            }
        }
    }
}

##################################################
# Convert CIDR masks to dotted quads             #
##################################################
sub ipv4_cidr2msk($) {
    return
      join( ".", unpack 'CCCC', pack( "B*", ( "0" x shift ) . ( "1" x 32 ) ) );
}

sub detect_os {
    my $OPTS = shift;
    my $data = {};
    my $os_release = '/etc/os-release';
    my $centos_release = '/etc/centos-release';

    # root: cat /etc/version
    # 2.4.4-RELEASE
    # root: cat /etc/platform
    # pfSense
    # root: uname -a
    # FreeBSD pfSense.localdomain 11.2-RELEASE-p10 FreeBSD 11.2-RELEASE-p10 #9 4a32323423423(RELENG_2_4_4): Wed May 15 18:54:42 EDT 2019     root@buildbot1-nyi.netgate.com:/build/ce-crossbuild-244/obj/amd64/ZfGpH5cd/build/ce-crossbuild-244/pfSense/tmp/FreeBSD-src/sys/pfSense  amd64
    my $etc_platform = '/etc/platform'; # pfsense stores platform 'pfsense'
    my $etc_version = '/etc/version';
    # detect debian
    if (-f "$os_release") {
        loadConfFileData($os_release, $data);

        &printMsg('DEBUG', "os data" . Dumper $data);

        $OPTS->{'os'} = lc $data->{'NAME'};
        $OPTS->{'os_full'} = $data->{'PRETTY_NAME'};
        $OPTS->{'os_version'} = $data->{'VERSION_ID'} if $data->{'VERSION_ID'};
        unless ($OPTS->{'os_version'}) {
            $OPTS->{'os_version'} = $data->{'VERSION'};
            $OPTS->{'os_version'} = $data->{'PRETTY_NAME'} unless $OPTS->{'os_version'};
            $OPTS->{'os_version'} =~ s/[^\d\.\-]//g;
        }
        $OPTS->{'os_base'} = 'debian'; # default
        if ($data->{'ID_LIKE'}) {
            $OPTS->{'os_base'} = (split " ", $data->{'ID_LIKE'})[0];
        } else {
            $OPTS->{'os_base'} = 'redhat' if ( $OPTS->{'os_full'} =~ /redhat/im);
            $OPTS->{'os_base'} = 'redhat' if ( $OPTS->{'os_full'} =~ /fedora/im);
            $OPTS->{'os_base'} = 'redhat' if ( $OPTS->{'os_full'} =~ /red hat/im);
            $OPTS->{'os_base'} = 'redhat' if ( $OPTS->{'os_full'} =~ /centos/im);
            $OPTS->{'os_base'} = 'debian' if ( $OPTS->{'os_full'} =~ /debian/im);
            $OPTS->{'os_base'} = 'debian' if ( $OPTS->{'os_full'} =~ /buntu/im);
        }
        $OPTS->{'os_base'} = lc $OPTS->{'os_base'};
    } # detect centos
    elsif(-f "$centos_release") {
        $OPTS->{'os_full'} = `cat $centos_release`;
        chomp($OPTS->{'os_full'});
        $OPTS->{'os'} = $OPTS->{'os_full'};
        $OPTS->{'os'} =~ s/\s?(linux|release)?\s?\d.+$//g;
        $OPTS->{'os_base'} = 'redhat' if ( $OPTS->{'os_full'} =~ /redhat/im);
        $OPTS->{'os_base'} = 'redhat' if ( $OPTS->{'os_full'} =~ /fedora/im);
        $OPTS->{'os_base'} = 'redhat' if ( $OPTS->{'os_full'} =~ /red hat/im);
        $OPTS->{'os_base'} = 'redhat' if ( $OPTS->{'os_full'} =~ /centos/im);
        $OPTS->{'os_base'} = 'debian' if ( $OPTS->{'os_full'} =~ /debian/im);
        $OPTS->{'os_base'} = 'debian' if ( $OPTS->{'os_full'} =~ /buntu/im);
        &printMsg('DEBUG', "os_full dump" . Dumper $OPTS->{os_full});
        &printMsg('DEBUG', "os_base dump" . Dumper $OPTS->{os_base});
        &printMsg('DEBUG', "os dump" . Dumper $OPTS->{os});
    } # detect PFsense
    elsif(-f "$etc_platform") {
        my $etc_platform = `cat $etc_platform`;
        chomp($etc_platform);
        $OPTS->{'os'} = `uname -o`;
        chomp($OPTS->{'os'});
        my $platform_version = '';
        if (-f "$etc_version") {
            $platform_version = `cat $etc_version`;
            chomp($platform_version);
        }
        my $uname_all = `uname -a`;
        chomp($uname_all);
        $OPTS->{'os_base'} = 'freebsd' if ( lc $uname_all =~ /freebsd/im);
        $OPTS->{'os_base'} = 'openbsd' if ( lc $uname_all =~ /openbsd/im);
        $OPTS->{'os_full'} = $OPTS->{'os'} . ' ' . $etc_platform . ' ' . $platform_version;
        &printMsg('DEBUG', "os dump" . Dumper $OPTS->{os});
        &printMsg('DEBUG', "os_full dump" . Dumper $OPTS->{os_full});
        &printMsg('DEBUG', "os_base dump" . Dumper $OPTS->{os_base});
    } else {
        $OPTS->{'os_full'} = 'Unknown';
        $OPTS->{'os_full'} = `cat /proc/version`;
        chomp($OPTS->{'os_full'});
        $OPTS->{'os'} = $OPTS->{'os_full'};
        $OPTS->{'os'} =~ s/\s?(linux|release)?\s?\d.+$//g;
        $OPTS->{'os_base'} = 'redhat' if ( $OPTS->{'os_full'} =~ /redhat/im);
        $OPTS->{'os_base'} = 'redhat' if ( $OPTS->{'os_full'} =~ /fedora/im);
        $OPTS->{'os_base'} = 'redhat' if ( $OPTS->{'os_full'} =~ /red hat/im);
        $OPTS->{'os_base'} = 'redhat' if ( $OPTS->{'os_full'} =~ /centos/im);
        $OPTS->{'os_base'} = 'debian' if ( $OPTS->{'os_full'} =~ /debian/im);
        $OPTS->{'os_base'} = 'debian' if ( $OPTS->{'os_full'} =~ /buntu/im);

    }

    my $platform = qx(uname -m);
    chomp($platform);
    $OPTS->{'arch'}     = 64 if ( $platform =~ /x86_64/ or $platform =~ /amd64/ );
    $OPTS->{'arch'}     = 32 if ( $platform =~ /i686/ or $platform =~ /i386/ );
    $OPTS->{'os_platform'} = $platform;
    &printMsg( "DEBUG", "Detected - [$OPTS->{'os'}] based on => [$OPTS->{'os_base'}] x$OPTS->{'arch'} platform => [$OPTS->{'os_platform'}] os_full => [$OPTS->{'os_full'}]" );

}

##################################################
# os_detecter : Detects OS type                  #
##################################################
sub osPlatformDetecter() {

    my $res      = shift;
    my $res_full = shift;
    my $platform;

    $platform = qx(uname -m);
    $$res     = 64 if ( $platform =~ /x86_64/ or $platform =~ /amd64/ );
    $$res     = 32 if ( $platform =~ /i686/ );

    chomp($platform);
    $$res_full = $platform;

    return 0;
}

############################################################
# osDestributionDetector : Detects OS version for Linux os #
############################################################
sub osDestributionDetector() {

    my $ver_ref = shift;

    eval {
        open( VFD, "/proc/version" )
          or die "Cannot determine Linux Distribution\n";
    };

    if ($@) {
        &printMsg( "ERROR", "$@" );
        return 1;
    }

    $_ = join( '', <VFD> );

    $$ver_ref = 'Unknown';
    $$ver_ref = 'RedHat' if (/Red Hat/m);
    $$ver_ref = 'Debian' if (/Debian/m);
    $$ver_ref = 'Ubuntu' if (/Ubuntu/m);

    &printMsg( "DEBUG", "Detected - [$$ver_ref]" );

    return 0;
}


##################################################
# calculateElapsedTime :                         #
##################################################
sub calculateElapsedTime() {

    my $start_epoch      = shift;
    my $elapsed_time_ref = shift;
    my ( $elapsed_time, $end_epoch );
    my ( $hr, $min, $sec );

    $end_epoch    = time();
    $elapsed_time = $end_epoch - $start_epoch;

    $sec = $elapsed_time % 60;
    $min = ( $elapsed_time / 60 ) % 60;
    $hr  = ( $elapsed_time / ( 60 * 60 ) ) % 24;

    $elapsed_time = sprintf( "%02d:%02d:%02d", $hr, $min, $sec );
    $$elapsed_time_ref = $elapsed_time;

    return 0;
}

##################################################
# checkDNSaccess : Verifies DNS connection       #
##################################################
sub checkDNSaccess() {

    my $dig     = shift;
    my $dnslist = shift;
    my $dnsok   = shift;
    my @dnslist = split( / /, $dnslist );
    my $dnsport = shift;
    my $cmd;

    &printMsg( "INFO", "Checking DNS servers access ..." );

    #If no dns list was received
    if ( !@dnslist ) {
        &printMsg( "ERROR", "DNS servers list is not defined." );
        return 1;
    }

    if ( !-e $dig ) {
        &printMsg( 'ERROR', "Missing dig [$dig] tool file." );
        return 1;
    }

    #Pass over dns servers list and try to connect
    foreach my $dns (@dnslist) {

        $cmd = "$dig -p $dnsport ptr \@$dns basic.threatstop.local 2>&1";
        &printMsg( 'INFO',  "Verifying dns server [$dns] access." );
        &printMsg( 'DEBUG', "$cmd" );
        my $rc = qx($cmd);
        my $rv = $?;

        if ( $rv != 0 ) {

            #Parse dig output and extract error message
            chomp($rc);
            my @rc = split( '\n', $rc );
            @rc = grep { /;; (?!global options)/ } @rc;
            $rc = join( ' ', @rc );
            $rc =~ s/;//g;
            chomp($rc);

            &printMsg( 'WARN', "dig command failed with error [$rc]" );
            next;
        }

        if ( $rc =~
            m/opcode: QUERY, status: NOERROR,.*QUERY: 1, ANSWER: (\d+)/s )
        {
            &printMsg( 'INFO', "$dns resolves OK." );
            push( @$dnsok, $dns );
        }
        else {
            &printMsg( "WARN",
"ThreatSTOP initial dig to $dns failed, skipping to the next server in the list."
            );
        }
    }

    if ( !defined @$dnsok[0] ) {
        &printMsg( "ERROR", "All ThreatSTOP DNS servers are unreachable." );
        return 1;
    }

    &printMsg( "INFO", "Done checking DNS servers." );
    return 0;
}

##########################################################
# loadConfigFileData : Load previous setup configuration #
##########################################################
sub loadConfFileData() {

    my $confile   = shift;
    my $data_hash = shift;
    my $err_msg   = shift;
    my $FILE;
    my $basedir = $TSDIR;

    #Reading configuration file
    &printMsg( "DEBUG", "Initializing data from configuration file." );

    if ( -e $confile ) {

        eval { open( FILE, "<", $confile ) or die "$!"; };
    }
    else {
        eval {
            open( FILE, "<", "$basedir/$confile" ) or die "$!";
            $confile = "$basedir/$confile";
        };
    }

    if ($@) {
        $$err_msg = "$@";
        return 1;
    }

    #Parse data
    while (<FILE>) {
        next unless (/^([A-Za-z\-\_]+)=[("]?(.*?)[")]?\s+$/);
        # &printMsg('DEBUG', "Added {$1} => '$2'");
        $data_hash->{$1} = $2;
    }
    close FILE;
    mpw( $confile, $data_hash );

    #Also update DNS servers list with pre-stored setting
    $DNS_LIST = $dns_server if ( defined $dns_server );

    return 0;
}

############################################################
# appendUploadLogEntry : Append an entry to the Upload Log #
############################################################
sub appendUploadLogEntry() {

    my $orig_filename    = shift;
    my $chunked_filename = shift;
    my $status_code      = shift;
    my $transfer_time    = shift;

    my $status     = "";
    my $status_msg = "";
    chomp( my $timestamp = `date -Iseconds` );
    my $FILE;

    if ($status_code == 0) {
        $status     = "SUCCESS";
        $status_msg = "Transfer time (seconds): ${transfer_time}"
    }
    elsif ($status_code == 1) { $status = "FAIL"; $status_msg = "Connection failed"; }
    elsif ($status_code == 2) {
        $status     = "FAIL";
        $status_msg = "Connection established, but send status was not OK";
    }
    else {
        $status     = "UNKNOWN";
        $status_msg = "Unrecognized status code [$status_code]";
    }

    open( FILE, ">>", "${UPLOADSLOGFILE}${curr_date}.log" ) or die "$!";
    print FILE
"$timestamp   $orig_filename   $chunked_filename   $status   $status_msg\n";
    close FILE;

    return 0;
}

##################################################################
# purgeOldUploadsLogs : Removes logs that are over ~2 months old #
#
# NOTE:  It would probably be better to implement this as a cron
#        instead of in ThreatStop.pm.  The setup routine could
#        be modified to accomplish this task.
##################################################################
sub purgeOldUploadsLogs() {

    my $logsdir = $TSDIR;

    my $rv =
      `find $logsdir -name 'tsuploads??????.log' -mtime +63 -exec rm -f {} \\;`;

    return 0;
}

##################################################################
# querySRV
#
# Using the passed service, protocol and domain, this will
# load the passed array reference as follows :
#
# args :
#   service name (without the underscore)
#   protocol name (without the underscore)
#   domain name to which the query should be made
#   reference to an array, which will get loaded as follows :
#
# $ref = [
#          { # record 0
#              priority => ...,
#              weight => ...,
#              target => ...,
#              port => ...,
#          },
#          { # record 1
#          ...
#          },
#        ]
#
# returns :
# 1 for success
# 0 for failure, which could mean bad input, or unable to query,
#   or empty return
#
sub querySRV {

    my $dig      = shift;
    my $service  = shift;
    my $protocol = shift;
    my $domain   = shift;
    my $arrayref = shift;

    &printMsg( "DEBUG", "SRV query for '_$service._$protocol.$domain'" );

    my $srvs = `$dig +tcp +short -t srv _$service._$protocol.$domain 2>&1`;

    my $status = $?;

    &printMsg( "DEBUG", "SRV query status ($status) return : $srvs" );
    if ( $status != 0 ) {

        return 0;
    }

    my ( $priority, $weight, $port, $target ) = ();
    foreach my $srv ( split( /\n/, $srvs ) ) {

        &printMsg( "DEBUG", "working on line : '$srv'" );

        if ( $srv =~ /^\s+$/ ) {

            &printMsg( "DEBUG", "Got an empty SRV '$srv'; skipping" );
            next;
        }

        ( $priority, $weight, $port, $target ) = split( /\s+/, $srv );

        &printMsg( "DEBUG", "SRV : $priority, $weight, $target, $port" );
        push @$arrayref,
          {
            priority => $priority,
            service  => $service,
            port     => $port,
            target   => $target,
          };
    }

    return 0 if !scalar @$arrayref;
    return 1;
}

my $COLOR_DEFS = {
    'normal'  => '',
    'black'   => "\033[1;30m",
    'grey'    => "\033[0;30m",
    'gray'    => "\033[0;30m",
    'red'     => "\033[0;31m",
    'bold_red'     => "\033[1;31m",
    'green'   => "\033[0;32m",
    'bold_green'   => "\033[1;32m",
    'yellow'  => "\033[0;33m",
    'bold_yellow'  => "\033[1;33m",
    'blue'    => "\033[0;34m",
    'bold_blue'    => "\033[1;34m",
    'magenta' => "\033[0;35m",
    'bold_magenta' => "\033[1;35m",
    'cyan'    => "\033[0;36m",
    'bold_cyan'    => "\033[1;36m",
    'white'   => "\033[0;37m",
    'bold_white'   => "\033[1;37m",
    'okblue'  => "\033[94m",
    'okgreen' => "\033[92m",
};

my $STYLE_DEFS = {
    'normal' => "\033[;0m",
    'invisible' => "\033[;8m",
    'reverse'   => "\033[;7m",
    'reset'     => "\033[0;0m",
    'bold'      => "\033[;1m",
    'underline' => "\033[4m",
    'header'    => "\033[95m",
    'fail'      => "\033[91m",
    'endc'      => "\033[0m",
    'warning'   => "\033[93m",
};

=head3 print_colorized
  simple method to print colorized text via control codes to terminal.
    args:
        text, string
        color, string (red, blue, green)
        style, string (normal*, reverse, reset, bold, underline, header, fail, endc, warning)
    returns:
        nothing
=cut

sub print_colorized {
    my ( $text, $color, $style ) = @_;
    $color ||= 'normal';
    $style ||= 'normal';

    print( $STYLE_DEFS->{$style} . $COLOR_DEFS->{$color} . $text . "\033[0m" );
}

sub colorized {
    my ( $text, $color, $style ) = @_;
    $color ||= 'normal';
    $style ||= 'normal';

    return $STYLE_DEFS->{$style} . $COLOR_DEFS->{$color} . $text . "\033[0m";
}

sub round {
    return sprintf "%.0f", shift;
}
=head3 print_banner

    same as print_colorized, but we subtract len($msg) from 80 / 2 to pad the
    output right / left and reverse the colors making a banner.
=cut
sub print_banner {
    my ( $text, $color, $style ) = @_;
    $color = 'bold_red' unless $color;
    $style = 'reverse' unless $style;
    # &printMsg('DEBUG', "color: $color style: $style");
    my $len = length($text);
    my $total_padding = 80 - $len;
    my $lr_padding = $total_padding > 0 ? round($total_padding / 2) : 0;
    print_colorized(' ' x $lr_padding . "$text" . ' ' x $lr_padding . "\n", $color, $style)
}


=head3 confirm_prompt

   reusable prompt that requires valid choices.
   args:
       msg - string with what they user will be confirming
       default - [0 or n = No, 1 or y = Yes]
   returns:
       0 - does not confirm
       1 - confirms
=cut

sub confirm_prompt {
    my ($msg, $default) = @_;
    my $default_output = '';
    if ($default and $default = lc($default) and $default =~ m/^y$|^yes$|^n$|^no$/) {
        $default_output = '[' . colorized(uc($default), 'cyan') . ']';
    } else {
        $default = '';
    }
    my $input;
    while (1) {
        print "\n $msg (y/n) $default_output";

        chomp( $input = lc <STDIN> );
        if ( $input =~ m/^$/ and $default ) {
            # user hit enter, took default and a default was supplied
            $input = $default;
        }
        if ( $input =~ m/^y$|^yes$/ ) {
            return 1;
        }
        elsif ( $input =~ m/^n$|^no$/ ) {
            return 0;
        }
        else {
            print_colorized( 'invalid choice', 'red' );
            next;
        }
    }
}

=head3 input_prompt

   reusable prompt that requires valid choices.
   args:
       msg - string with what they user will be confirming
       regex - matching regular expression
       default - default string
       silent - do not show output
       nullable - empty string allowed as value

   returns:
       0 - does not confirm
       1 - confirms
=cut

sub input_prompt {
    my ($msg, $regex, $default, $silent, $nullable) = @_;
    $default = 'none' if (! $default and $nullable );
    my $input;
    my $default_ans = $default ? '[' . colorized($default, 'cyan') . ']' : '';
    while (1) {
        print "\n $msg $default_ans ";
        chomp( $input = <STDIN> );
        if ($input eq '') {
            if ($default eq 'none') {
                print_colorized( "\nset: -empty-\n", 'green' ) unless $silent;
                return '';
            }
            elsif (defined $default and $default) {
                print_colorized( "\nset: $default\n", 'green' ) unless $silent;
                return $default;
            }
            else {
                print_colorized( 'invalid value', 'red' );
                next;
            }
        } elsif ( $input =~ m/$regex/ ) {
            # fix double quotes (can't have because of conf)

            print_colorized( "\nset: $input\n", 'green' ) unless $silent;
            return $input;
        }
        else {
            print_colorized( 'invalid value', 'red' );
            next;
        }
    }
}


sub get_backup_path {
    my $backup_file = shift;
    my $base = basename $backup_file;
    return "${TSDIR}/backups/${base}";
}

sub get_or_create_dir {
    my $dir = shift;

    unless ( -d $dir ) {
        printMsg( "INFO", "$dir does not exists, creating..." );
        qx(mkdir -p $dir 2> /dev/null);
        return $?;
    }
    return 0;
}

=head3 backup_file

    1. Creates a backup dir if not exists
    2. Copies file into backup dir with .orig appended unless exists
    3. Copies file into backup dir with .timestamp appended unless exists

    args:
        string - file path to back up
    returns:
        0 - success
        1 - failure

    WARNING: if you backup two different files named foo.conf you WILL have unexpected consequences

=cut
sub backup_file {
    my $filepath = shift;

    my $backup_dir = "${TSDIR}/backups";
    return 1 if (get_or_create_dir($backup_dir) != 0);

    my $orig = get_backup_path($filepath) . '.orig';
    my $version = get_backup_path($filepath) . '.' . time;

    my @backup_paths = ("$orig", "$version");
    for my $outfile (@backup_paths) {
        unless ( -f "$outfile" )
        {
            printMsg( "INFO", "Backing up $filepath to $outfile" );
            my $cmd = "$SUDO cp -p $filepath $outfile 2>&1 > /dev/null";
            &printMsg('DEBUG', "Running: $cmd");
            eval { system( $cmd ) == 0 or die "System call failed: $cmd $!"; };
            if ($@) {
                &printMsg('ERROR', "Running: $cmd ERROR: $@");
                return 1;
            }
        }
    }

    return 0;
}

sub restore_file {
    my $filepath = shift;
    my $orig = get_backup_path($filepath) . '.orig';

    if ( -f $orig ) {
        my $cmd = "$SUDO cp -p $orig $filepath 2>&1 > /dev/null";
        &printMsg('DEBUG', "Running: $cmd");
        eval { system( $cmd ) == 0 or die "System call failed: $cmd $!"; };
        if ($@) {
            &printMsg('ERROR', "Running: $cmd ERROR: $@");
            return 1;
        } else {
            &printMsg('INFO', "Restored '$filepath' successfully.");
        }
    } else {
        &printMsg('ERROR', "The backup of '${filepath}' was not found at '${orig}'. ");
        return 1;
    }
    #TODO: write a this to see if any with timestamp exist, and use the earliest ver if orig doesnt exist
    return 0;
}

=head3 add_or_remove_lines_in_file

reusable method for adding/removing a list of text to a file, we always create a backup of the file first.

0. create backup of original
1. write text to disk, if exists, loops through old file and removes first before proceeding
2. should be idempotent
3. takes string: /path/to/text
4. iterates block first to remove any lines that match, then adds them line by line.
5. stores the block of text in a file

args:
    string - path to system config file to upsert
    array - text lines

=cut

sub add_or_remove_lines_in_file {
    # print "Dump: " . Dumper @_;
    my $action = shift; # add | remove
    my $filename = shift;
    my $lines = shift;

    return 1 if ($action ne 'add' and $action ne 'remove');

    my $file_basename = basename $filename;

    return 1 if (-f "$filename") and (backup_file($filename) != 0);

    # exit early
    if (ref($lines) ne "ARRAY" and ref($lines) ne "ARRAYREF") {
        &printMsg('ERROR', "expected ARRAY of lines, got " . ref($lines));
        return 1;
    }
    if (scalar @{$lines} < 1) {
        &printMsg('ERROR', "expected at least one line to write to $filename");
        return 1;
    }
    if (! -f $filename) {
        return 1 if $action eq 'remove'; # we can't remove lines from a non-existent file
        &printMsg('DEBUG', "Creating file: $filename");
        eval { qx(touch $filename); };
        if ($@) {
            &printMsg('ERROR', "Error in creating '$filename' Err:" . $@);
            return 1;
        }
        &printMsg('OK', "Created file: $filename");
    }
    if (! -f $filename or ! -w $filename) {
        &printMsg('ERROR', "expected $filename exists and target conf writable by effective UID:GID");
        return 1;
    }


    my $in;
    my $out;
    my @data;
    eval {
        open( $in, '<', $filename );
        @data = <$in>;
        close($in);
        open( $out, '>', "${filename}.tmp" );
    };
    if ($@) {
        &printMsg( "ERROR", $@ );
        return 1;
    }
    my $missing = [];
    my $there = [];
    for my $line (@{$lines}) {
        &printMsg('DEBUG', "LINE: $line");
        if ( grep { chomp($_); $line eq $_ } @data ) {
            push @{$there}, $line;
            # do remove
        }
        else {
            # do remove
            push @{$missing}, $line;
        }
    }
    for my $line (@data) {
        if ($action eq 'remove') {
            print $out qq|$line\n| unless grep { $line eq $_ } @{$there};
        } else {
            print $out qq|$line\n|;
        }
    }
    if ($action eq 'add') {
        for my $line (@{$missing}) {
            print $out qq|$line\n|;
        }
    }
    # ADD - loop again and add line if not seen in file

    print "data: " . Dumper @data;
    print "there: " . Dumper $there;
    print "missing: " . Dumper $missing;
    close($out);
    if (-f "${filename}.tmp") {
        move("${filename}.tmp", "$filename") or die("failed moving ${filename}.tmp to $filename" );
    } else {
        &printMsg('OK', "Nothing to move");
    }
    &printMsg('OK', "wrote $filename");
    return 0;

}

sub file_to_array {
    my $filename = shift;
    unless (-f $filename) {
        &printMsg('FAIL', "$filename can not be read or located.");
        return 1;
    }
    my @data;
    my $in;
    eval {
        open( $in, '<', $filename );
        @data = <$in>;
        close($in);
    };
    if ($@) {
        &printMsg( "ERROR", $@ );
        return 1;
    }
    return @data;
}

#exported aliases
sub add_to_file {
    return add_or_remove_lines_in_file('add', @_);
}
sub remove_from_file {
    return add_or_remove_lines_in_file('remove', @_);
}

=head3 merge_system_config

This merges a file into a system config (like /etc/rc.local).

1. backup the original file
2. use '$SUDO cat >>' to append to the file.

=cut

### UPDATE: This function should be localized for in the LocalDevice module.
################################################################################
#install_rc_local : generates & installs rc.local ThreatSTOP config            #
################################################################################
# sub install_rc_local() {

#     &printMsg( "DEBUG", "Creating local copy of rc.local file." );

#     #Read original rc.local file
#     open( FILE, '</etc/rc.local' );
#     $_ = join( '', <FILE> );
#     my $ORGRCLOCAL = $_;
#     close FILE;

#     #Update ipsetget setting
#     if (/\S+ipsetget.pl/) {
#         s/\S+ipsetget.pl/${TSDIR}\/ipsetget.pl > \/tmp\/ipsetget.log\ 2>&1/s;
#     }
#     else {
#         s/^exit 0/${TSDIR}\/ipsetget.pl > \/tmp\/ipsetget.log\ 2>&1\nexit 0/sm;
#     }

#     if ( $_ eq $ORGRCLOCAL ) {
#         &printMsg( "DEBUG", "Skipping, as rc.local already contains the correct command\n" );
#         $ORGRCLOCAL = 1;
#     }
#     else {
#         eval { open( FILE, '>' . ${TSDIR} . '/rc.local' )
#               or die "$!"; };

#         if ($@) {
#             &printMsg( "ERROR", "Opening local rc.local failed with error [$@]" );
#             return 1;
#         }
#         print FILE $_;
#         close FILE;

#         eval { open( FILE, '>' . ${TSDIR} . '/rc.local.org' )
#               or die "$!"; };

#         if ($@) {
#             &printMsg( "ERROR", "Opening local rc.local.org failed with error [$@]" );
#             return 1;
#         }

#         print FILE $ORGRCLOCAL;
#         close FILE;

#         $ORGRCLOCAL = "";
#     }

#     &printMsg( "DEBUG", "Copying modified rc.local -> /etc/rc.local" );

#     if ( $ORGRCLOCAL eq '1' ) {
#         &printMsg( "DEBUG", "rc.local file is already up-to-date ... Skipping ..." );
#         return 0;
#     }

#     $cmd = "mv -f ${TSDIR}/rc.local /etc; chmod +x /etc/rc.local";
#     $cmd = "$SUDO $cmd" if ( get_UID() != 0 );

#     $rv = qx($SUDO $cmd 2>&1);
#     $rc = $?;

#     if ( $rc != 0 ) {
#         &printMsg( "ERROR", "rc.local deploy failed with error [$rv]" );
#         return 1;
#     }


#     return 0;
# }


=head3 line_in_config

    reusable function to see if a configuration line located in a config file

=cut

sub line_in_file {
    my $findline = shift;
    my $file     = shift;

    return 0 unless $findline;
    print "file does not exist" unless -f $file;
    printMsg( "DEBUG", "File: $file does not exist or is not readable." )
      unless -f $file;
    return 0 unless -f $file;

    open my $info, $file or die "Could not open $file: $!";

    while ( my $line = <$info> ) {

        # print $line;
        return 1 if $line =~ m/${findline}/;

        # last if $. == 30; # $. is line number
    }
    close $info;
    return 0;
}


##################################################################################################
#install_syslog_settings : deploy log settings from tmp files we have created before                   #
#                    logrotated.messages.tmp --> /etc/logrotate.d/messages                       #
#                    01-threatstop-rsyslog.conf.tmp --> /etc/rsyslog.d/01-threatstop-rsyslog.conf#
#                    threatstop-logrotate.conf --> /etc/logrotate.d/threatstop-logrotate.conf    #
#                    $TMP_LOGFILE1 = logrotated.messages.tmp                                     #
#                    $TMP_LOGFILE2 = 01-threatstop-rsyslog.conf.tmp                              #
#                    $TMP_LOGFILE3 = threatstop-logrotate.conf.tmp                               #
##################################################################################################
# setup logrotate, rsylog, syslog-ng etc.
sub install_syslog_settings() {

    my $rc;
    my $rv;

    die('no install_syslog_settings created yet');

    return 0;
}



sub get_UID {
    return $<
}


##################################################################
# abortIfNotUser : Don't allow script to run unless running as the
# passed user.
# arg : name of user to check
# returns : 0 for fail, 1 for success
##################################################################
sub abortIfNotUser {

    my $username = shift;
    my ($name, $passwd, $uid, @whocares) = getpwnam($username);
    if ( $> != $uid ) {
        &printMsg( "ALERT", "$0 needs to be run as $username." );
        exit 1;
    }

    return 0;
}


=head4 check_for_app

Generic check for app tries to identify app's location on disk.

args:
    string - app basename
    string - is_fatal (0 = no, any positive = yes)
    hash ref - $OPTS

returns:
    0 - ok
    1 - fail
=cut
sub check_for_app {
    my $name = shift;
    my $is_fatal = shift; # any positive value signals abortExec if missing
    my $OPTS = shift;

    # exit early if already set and exists and executable
    if (exists $OPTS->{$name} and -f $OPTS->{$name} and -x $OPTS->{$name}) {
        &printMsg('CHECK', 'pre-provided location looks good: ' . $OPTS->{$name});
        return 0;
    }

    my @possible_dirs = ( '/bin/', '/usr/sbin', '/usr/bin', '/sbin', '/usr/local/sbin', '/usr/local/bin', '/opt/bin', '/opt/sbin' );

    my $exit;

    my $location;
    eval {
        $location = qx($SUDO which $name 2> /dev/null);
        $exit = $?;
    };
    unless($location and $exit == 0) {
        $location = qx(which $name 2> /dev/null);
        $exit = $?;
    }
    unless($location and $exit == 0) {
        for my $dir (@possible_dirs) {
            my $fullpath = "$dir/$name";
            if ( -f $fullpath and -x $fullpath ) {
                $location = "$fullpath";
                last;
            }
        }
    }
    if ($location) {
        chomp($location);
        &printMsg('WARN', "Updating $name app location $OPTS->{$name} => $location") if (exists $OPTS->{$name});
        &printMsg('CHECK', "Found $name @ [$location]");
    } else {
        &printMsg('DEBUG', "Could not find $name location");
        delete $OPTS->{$name};
        abortExec() if $is_fatal;
        return 1;
    }
    $OPTS->{$name} = $location;
    # print "DUMP: " . Dumper $OPTS;
    # exit 0;

    return 0;
}

sub get_package_manager_install_cmd {
    my $os = shift;
    return unless $os;
    $os = lc($os);
    if ( grep(/^${os}^/, qw(debian ubuntu raspbian lubuntu xubuntu kubuntu edubuntu) ) ) {
        return '/usr/bin/apt-get install -y';
    }
    elsif ( grep(/^${os}^/, qw(rhel centos redhat) ) ) {
        return '/usr/bin/yum -y';
    }
    elsif ( grep(/^${os}^/, qw(openbsd freebsd) ) ) {
        return 'env ASSUME_ALWAYS_YES=YES pkg';
    } else {
        &printMsg('WARN', "Sorry no package manager found for ${os}. ");
    }
}

# replaced with a hash
# sub get_package_name {
#     my $mod = shift;
#     my $os = shift;
#     return unless $os and $mod;
#     if ( ( $os eq 'Debian' ) || ( $os eq 'Ubuntu' ) ) {
#         return '/usr/bin/apt-get -y';
#     }
#     elsif ( $os eq 'RedHat') {

#         return $mod_to_package->{$mod}->{$os};

#     } else {
#         &printMsg('WARN', "Sorry no package manager found for ${os}. ");
#     }
# }

# at the moment only yum and apt are supported
sub install_via_package_manager {
    my $mod = shift;
    my $os = shift;
    my $OPTS = shift;

    check_for_app($mod, 0, $OPTS);
    if (exists $OPTS->{$mod}) {
        &printMsg('DEBUG', "Found $mod at $OPTS->{$mod}, no need to install");
        return 0;
    }

    &printMsg('DEBUG', "checking if system pkg-mgr has $mod for $os " . Dumper $OPTS);
    return 1 unless $os and $mod and $OPTS;

    my $pkgcmd = get_package_manager_install_cmd($os);
    my $pkgname = $mod;
    my $cmd = "$SUDO $pkgcmd $pkgname";
    print_colorized("Unable to locate ${mod}.", 'red');
    if ( confirm_prompt("Do you want to install '$mod' via '$cmd'?") ) {
        # do install and return exit status code 0 = success, > 1 = fail
        my $exit;
        if ( system($cmd) == 0) {
            $exit = $?;
            &printMsg('OK', "Installed app-pkg: $mod");
        } else {
            $exit = $?;
            &printMsg('FAIL', "Installed app-pkg: $mod");
        }
        return $exit;
    } else {
        # user chose not to install module.
        return 1;
    }
}

sub install_via_cpan {
    my $mod = shift;
    my $OPTS = shift;

    return 1 unless $OPTS and $mod;

    my $pkgcmd = $OPTS->{'cpanm'} ? $OPTS->{'cpanm'} : $OPTS->{'cpan'};
    my $cmd = "$pkgcmd $mod";
    if ( confirm_prompt("Unable to locate ${mod}. Do you want to install the module via '$cmd'?") ) {
        # do install and return exit status code 0 = success, > 1 = fail
        return system($cmd);
    } else {
        # user chose not to install module.
        return 1;
    }
}

=head4 check_for_module

Check for Perl Module

args:
    string - app basename
    string - is_fatal (0 = no, any positive = yes)
    hash ref - $OPTS

returns:
    0 - ok
    1 - fail
=cut
sub check_for_module {
    my $mod = shift;

    # chomp( my $lwp = qx($SUDO perl -MLWP::UserAgent -e 'print "FOUNDLWP";' 2>&1) );
    # if ( $lwp !~ /^FOUNDLWP$/ ) {
    #     &printMsg( "DEBUG", "load_environment() : LWP::UserAgent NOT found in perl's library search path'" );
    #     return 1;
    # }
    # &printMsg( "DEBUG", "LWP::UserAgent [checked]" );
    my $rc = eval( 'use ' . $mod . '; return "$mod Version: $' . $mod . '::VERSION;\n"' );

    # if we don't have it installed and is_fatal is true:
    if ($@) {
        return 1;
        &printMsg( "DEBUG", "$mod => [not found]" );
    };
    &printMsg( "DEBUG", "$mod => [found] Version: $rc" );
    return 0;
}

sub get_app_version {
    my $app = shift;
    unless ($app) {
        &printMsg('DEBUG', "get_app_version was not passed \$app");
        return;
    }
    my @version_flags = ('-v', '-V', '--version');
    for my $flag (@version_flags) {
        my $output;
        my $exit;
        eval {
            chomp($output = `$SUDO $app $flag 2>&1`);
            $exit = $?;
        };
        if ($@ or $exit != 0) {
            chomp($output = `$app $flag 2>&1`);
            $exit = $?;
        }
        # print 'output: ' . $output;
        # attempt to parse out version numbers https://regex101.com/r/tqBsjM/1
        if ($output =~ /\D?(\d+[\.-]\d+(?:[\.-]\d+)?)/sg) {
            my $version = $1;
            &printMsg('CHECK', "$app parsed version [$version]");
            return $version;
        } elsif ( $exit == 0 ) {
            &printMsg('DEBUG', "$app $flag exitted OK, but could not parse version from output: $output");
            return 1;
        }
    }
}

sub app_is_running {
    my $app = shift;
    my $results = shift || 'no';

    my $output = qx(ps -ax | grep -v grep | grep $app);
    if ($? != 0 and $output =~ m/BusyBox/i) {
        $output = qx(ps -w | grep -v grep | grep $app);
    }

    if (defined $results and $results ne 'no') {
        $$results = $output;
    }
    return 0 if $output;
    return 1;
}

sub find_app_confd {
    my $app = shift;
    my $results = shift || 'no';

    # shortcut to known location if exists
    my $known_conf_dir_by_app = {
        logrotate => ['/etc/logrotate.d', '/usr/local/etc/logrotate.d'],
        rsyslog => ['/etc/rsyslog.d'],
        cron => ['/etc/cron.d', '/etc/crontab.d', '/opt/etc/cron.d', '/usr/local/etc/cron.d'],
        crontab => ['/etc/cron.d', '/etc/crontab.d', '/opt/etc/cron.d'],
        'syslog-ng' => ['/etc/syslog-ng/conf.d', '/opt/etc/syslog-ng.d' ],
        'stunnel' => ['/etc/stunnel', '/usr/local/etc/stunnel/conf.d', '/opt/threatstop' ],
        'stunnel4' => ['/etc/stunnel', '/usr/local/etc/stunnel/conf.d', '/opt/threatstop' ],
    };
    if (exists $known_conf_dir_by_app->{$app}) {
        for my $dir (@{$known_conf_dir_by_app->{$app}}) {
            if (-d $dir) {
                if (defined $results and $results ne 'no') {
                    $$results = $dir;
                    return 0;
                }
            }
        }
    }

    # else try longer heavier method
    my @config_dirs = ('/etc', '/opt/etc', '/usr/local/bin/etc', '/usr/local/etc', '/usr/share/etc');
    my $output;

    for my $cdir (@config_dirs) {
        if (-d $cdir) {
            if (exists $known_conf_dir_by_app->{$app}) {
                for my $dir (@{$known_conf_dir_by_app->{$app}}) {
                    my $dir_base = basename $dir;
                    $output = qx( find $cdir -type d -name "$dir_base" 2> /dev/null | grep $app);
                    chomp($output);
                    last if $output;
                }
            }
        }
    }


    if (defined $results and $results ne 'no') {
        $$results = $output;
    }
    return 0 if $output;
    return 1;
}

=head3 find_app_conf

NOTE: This is to be used only if confd is not available on the system.
NOTE: We much more prefer to cleanly extend our configuration as a conf.d/ts.conf file

Try to locate a system applications configuration file in /etc, then others if that doesn't pan out.



=cut
sub find_app_conf {
    my $app = shift;
    my $results = shift || 'no';

    # shortcut to known location if exists
    my $known_conf_dir_by_app = {
        rsyslog => ['/etc/rsyslog.conf'],
        cron => [ '/etc/crontab', '/opt/etc/crontab' ],
        crontab => [ '/etc/crontab', '/opt/etc/crontab'],
        'syslog-ng' => [ '/etc/syslog-ng/syslog-ng.conf', '/opt/etc/syslog-ng/syslog-ng.conf' ],
    };
    if (exists $known_conf_dir_by_app->{$app}) {
        for my $dir (@{$known_conf_dir_by_app->{$app}}) {
            if (-f $dir) {
                if (defined $results and $results ne 'no') {
                    $$results = $dir;
                    return 0;
                }
            }
        }
    }

    # else try longer heavier method
    my @config_dirs = ('/etc', '/opt/etc', '/usr/local/bin/etc', '/usr/share/etc');
    my $output;

    for my $cdir (@config_dirs) {
        if (-d $cdir) {
            $output = qx( find $cdir -type f -name "${app}*conf" 2> /dev/null | grep -v init);
            chomp($output);
            last if $output;
        }
    }

    if (defined $results and $results ne 'no') {
        $$results = $output;
    }
    return 0 if $output;
    return 1;
}


=head3 detect_app_info

We should basically look for (rsyslog, syslog-ng, syslog) else syslog_parser and ask for syslog location.

1. First check if we have rsyslog installed (seems to be the default for debian)
 - no
2. next, check if we have syslog-ng installed
3. finally, default to a syslog_parser (TBDeveloped)

# confirm logger settings
1a. to see if we have x installed, check for executable binary running via ps -augx|grep -v grep|grep x
 - yes
2a. if we do make sure we have the corresponding /etc/conf.d: find /etc -type d -name "rsyslog*" 2> /dev/null
 - no
2b. check if we have /etc/x.conf # don't modify directly, instead we'll generate x conf and have user integrate it.
   i.e. "We didn't detect /etc/rsyslog.d to install the generated logging configuration. Please integrate this in your systems logger"

note: keep in mind we'll have to know how to handle restarting different types of loggers

set config dir $OPTS->{'logger_confd'} = '/etc/rsyslog.d'
finally set logger type $OPTS->{'logger'} = 'rsyslog'
we'll use logger value to generate_logger_restart_cmd function

args:
    hash - $OPTS
    string - generic app type (logger, scheduler, rotater, etc...)
        app type used in $opts->{$app_type} to make is callable no matter what app is used
    array - supported application basenames

=cut


sub detect_app_info {
    my ($OPTS, $app_type, @supported_executables) = @_;


    my $app_confd;
    my $config;
    for my $app ( @supported_executables ) {
        &printMsg('START', "Detecting $app");
        # try to benefit from earlier application environment detection
        if ( exists $OPTS->{$app} or exists $OPTS->{"${app}d"} ) {
            if (find_app_confd($app, \$app_confd) == 0) {
                $OPTS->{"$app_type"} = $app;
                $OPTS->{"${app_type}_confd"} = $app_confd;
                &printMsg('CHECK', "Found $app and conf dir [$app_confd]");
                return 0 if $app and $app_confd;
            } elsif((find_app_conf($app, \$config) == 0)) {
                $OPTS->{"$app_type"} = $app;
                $OPTS->{"${app_type}_conf"} = $config;
                &printMsg('CHECK', "Found $app and conf file [$config]");
                return 0 if $app and $config;
            }

        }
        # else go to running detection
        if (app_is_running($app) == 0) {
            &printMsg('START', "Detecting $app conf dir");
            if (find_app_confd($app, \$app_confd) == 0) {
                $OPTS->{"$app_type"} = $app;
                $OPTS->{"${app_type}_confd"} = $app_confd;
                &printMsg('CHECK', "Found $app and conf dir [$app_confd]");
                return 0 if $app and $app_confd;
            } elsif((find_app_conf($app, \$config) == 0)) {
                $OPTS->{"$app_type"} = $app;
                $OPTS->{"${app_type}_conf"} = $config;
                &printMsg('CHECK', "Found $app and conf file [$config]");
                return 0 if $app and $config;
            }

        }
    }
    return 1;
}


=head3 get_app_restart

detect & return app (rsyslog, syslog-ng, logrotate) restart command (systemd | sysv)

    args:
        hash - $OPTS
    returns:
        string - command


=cut

sub get_app_restart {
    my ($OPTS, $app) = @_;


    my $exit = 100;
    my $service_status;
    my $service_cmd;
    $service_cmd = $OPTS->{service} if exists $OPTS->{service};

    if (defined $service_cmd and $service_cmd) {
        my $cmd = "$service_cmd $app";
        $service_status = `$cmd status 2>&1`;
        &printMsg('DEBUG', "running '$cmd' to see if systemD service is running: $service_status");
        $exit = $?;
        if ($service_status =~ m/insufficient privileges/i) {
            $service_status = `$SUDO $cmd status 2>&1`;
            &printMsg('DEBUG', "re-running '$SUDO $cmd' to see if systemD service is running: $service_status");
            $exit = $?;
        }
    }


    # check for systemd
    if (defined $service_status and $service_status and $exit == 0) {
        my $cmd = "$service_cmd $app";
        my $service_usage = `$cmd 2>&1`;
        &printMsg('DEBUG', "running '$cmd' to parse out restart options: $service_usage");
        if ($service_usage =~ m/[^-]restart/) {
            return "$cmd restart";
        }
        elsif ($service_usage =~ m/[^-]rotate/) {
            return "$cmd rotate";
        }
        elsif ($service_usage =~ m/[^-]reload/) {
            return "$cmd reload";
        }
    }
    if ( -f "/etc/init.d/$app" ) {
        # detect systemV style
        &printMsg('DEBUG', "detected sys-v for $app");
        my $cmd = "/etc/init.d/$app";
        my $service_usage = `$cmd 2>&1`;
        &printMsg('DEBUG', "running '$cmd' to parse out restart options: $service_usage");
        if ($service_usage =~ m/[^-]rotate/) {
            return "$cmd rotate";
        }
        elsif ($service_usage =~ m/[^-]reload/) {
            return "$cmd reload";
        }
        elsif ($service_usage =~ m/restart/) {
            return "$cmd restart";
        }
    }
    else {

        if ( exists $OPTS->{systemctl} ) {
            my $cmd = "$OPTS->{systemctl} status $app";
            my $service_usage = `$cmd 2>&1`;
            $exit = $?;
            &printMsg('DEBUG', "running '$cmd'[$exit] output: $service_usage");
            if ($service_usage =~ m/active/i) {
                &printMsg('DEBUG', "detected '$OPTS->{systemctl} restart $app'");
                return "$OPTS->{systemctl} restart $app";
            }
        }
        # last ditch effor, look for a glob of the logger in case
        # of /etc/init.d/04_rsyslogd for rsyslog etc..
        my @glob_loggers = glob("/etc/init.d/*$app*");
        for my $cmd (@glob_loggers) {
            my $service_usage = `$cmd 2>&1`;
            if ($service_usage =~ m/[^-]rotate/) {
                return "$cmd rotate";
            }
            elsif ($service_usage =~ m/[^-]reload/) {
                return "$cmd reload";
            }
            elsif ($service_usage =~ m/restart/) {
                return "$cmd restart";
            }
        }
        &printMsg( "DEBUG", " *** WARNING *** no $app restart command detected." );
        return '';
    }
    return 0;
}

sub install_app_config {
    my ($OPTS, $app, $template, $search_and_replace, $chown, $mode) = @_;
    # $chown = 'root:root' unless $chown; # don't assume we want to chown
    # the file in confd method
    my $executable = (exists $OPTS->{$app} and $OPTS->{$app}) ? $OPTS->{$app} : '';
    unless ($executable) {
        # exit early
        &printMsg('DEBUG', "Refusing to attempt installation without a path to $app executable: '$executable' defined");
        return 1;
    }

    my $app_restart = (exists $OPTS->{"${app}_restart_cmd"} and $OPTS->{"${app}_restart_cmd"}) ? $OPTS->{"${app}_restart_cmd"} : '';
    my $conf_dir = $OPTS->{"${app}_confd"} || '';
    my $config = $OPTS->{"${app}_conf"} || '';
    # my $template = "$OPTS->{src}/templates/app.d/threatstop_iptables.conf";
    my $state = {}; # hash we will use to store installed logging state.
    my $threatstop_conf = "$OPTS->{src}/threatstop.conf";
    my $source_to_target_paths;
    unless (-f $template ) {
        &printMsg('ERROR', "No $app source configuration file: '$template' found. Required to install app $app.");
        return 1;
    }
    if (defined $conf_dir and $conf_dir) {
        # Install via conf.d method (best option)
        # copy template and search & replace file
        &printMsg('DEBUG', "Installing $template into $conf_dir");
        my $target_path = $conf_dir . '/' . basename $template;
        if (-f $target_path) {
            &printMsg('DEBUG', "Removing previous threatstop config: $target_path ");
            qx($SUDO rm $target_path);
        }
        $source_to_target_paths = { $template => $target_path };
        # $search_and_replace = { '__TSLOGFILE__' => $OPTS->{logfile} }; # provided in args
        # take mode, owner defaults
        if (install_config($source_to_target_paths, $search_and_replace, $chown) == 0) {
            # set the path our app config
            $OPTS->{"installed_${app}_config"} = $target_path;
            &printMsg('DEBUG', "Updating $threatstop_conf: " . Dumper $OPTS->{"installed_${app}_config"});
            write_conf_key($threatstop_conf,  { "installed_${app}_config" => $target_path});
        } else {
            # we failed installing the configuration file.
            &printMsg('FAIL', "could not install $app app");
            return 1;
        }
    }
    elsif (defined $config and $config) {
        # install via merge into app's .conf file (riskier :/)
        # we are merging into main app config
        &printMsg('START', "Merging into ${app}'s config: $config");
        my $config = [];
        # read in source template config into array
        @$config = file_to_array("$template");
        &printMsg('DEBUG',  "read in file into array: " . Dumper $config);
        return 1 if (get_or_create_dir("$OPTS->{src}/deploy") != 0);
        my $config_basename = basename $config;
        my $merge_file = "$OPTS->{src}/deploy/$config_basename";
        # backup original conf
        return 1 if (backup_file($config) != 0); # if we have a default "parse_app" script remove the early exit

        # create a file with config section only
        if (-f $merge_file) {
            &printMsg('DEBUG', "Removing previous threatstop merge config: $merge_file ");
            qx($SUDO rm $merge_file);
        }
        $source_to_target_paths = { $template => $merge_file };
        # $search_and_replace = { '__TSLOGFILE__' => $OPTS->{logfile} }; # provided in args
        unless (install_config($source_to_target_paths, $search_and_replace, $chown) != 0) {
            &printMsg('ERROR', "Could not stage '$template' at '$merge_file' for config merging. $@");
            return 1;
        }

        # copy config to staging area
        eval { qx($SUDO cp -p $config ${merge_file}.tmp) or die("Could not copy & stage $config to $merge_file for config merging."); };
        if ($@) {&printMsg('ERROR', $@); return 1; }
        # append our config section to config and swap it back, clobbering config
        eval { qx($SUDO cat ${merge_file} >> ${merge_file}.tmp) or die("Could not merge our config section into ${merge_file}.tmp"); };
        if ($@) {&printMsg('ERROR', $@); return 1; }
        # swap back.
        eval { qx($SUDO cp -p ${merge_file}.tmp $config) or die("Could not install merged $config"); };
        if ($@) {&printMsg('ERROR', $@); return 1; }

        # set installed_app breadcrumbs for uninstall
        # use a hash delta config section file => installed config
        $OPTS->{"installed_${app}_config"} = $config;
        &printMsg('DEBUG', "Updating $threatstop_conf: " . Dumper $OPTS->{"installed_${app}_config"});
        write_conf_key($threatstop_conf,  { "installed_${app}_config" => $config});

        $OPTS->{"installed_${app}_merge"} = $merge_file;
        &printMsg('DEBUG', "Updating $threatstop_conf: " . Dumper $OPTS->{"installed_${app}_merge"});
        write_conf_key($threatstop_conf,  { "installed_${app}_merge" => $merge_file});
        # write_conf_keys($threatstop_conf, $threatstop_conf, $state);



    } else {
        # no main app method.
        #TODO: Maybe here we default to our custom "parse_app"
        &printMsg('ERROR', "No installation path found for $app");
        return 1;
    }

    if ($app_restart) {
        eval {
            `$SUDO $app_restart` ? &printMsg('OK', "restarted $app via $app_restart") : &printMsg('FAIL', "restarted $app via $app_restart");
        };
    } else {
        &printMsg('DEBUG', "No $app restart command defined, skipping restarting app");
    }


    return 0;
}

sub get_file_stat {
    my $filename = shift;

    my $info = {};
    my @ftypes;

    if ( my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = lstat($filename) ) {
        $info->{uid} = $uid;
        $info->{user} = getpwuid($uid);
        $info->{group} = getgrgid($gid);
        $info->{gid} = $gid;
        $ftypes[S_IFDIR] = "d";
        $ftypes[S_IFCHR] = "c";
        $ftypes[S_IFBLK] = "b";
        $ftypes[S_IFREG] = "-"; # regular file
        $ftypes[S_IFIFO] = "p";
        $ftypes[S_IFLNK] = "l";
        $ftypes[S_IFSOCK] = "s";
        $info->{mode} = sprintf "%04o", S_IMODE($mode);
        my $filetype = S_IFMT($mode);
        $info->{type} = $ftypes[$filetype];
        return $info;
    } else {
            print "File does not exist or unreadable\n";
    }

}




=head3 install_config

    uses system call to perl -i.bak (creates .bak backup file) and generates a separate -pe for each replacement
    args:
        source_to_target_paths - Hash source => destination file path map
        translations - Hash search => replace key value pairs
        chown - Scalar user:group format
        mode - Int file posix chmod mode
    returns:
        0 for success
        1 for failure

=cut
sub install_config {
    my $source_to_target_paths = shift;
    my $translations = shift;
    my $chown = shift; # fetch from origin file if unset
    my $mode = shift; # fetch from origin file if unset

    my $owner;
    my $group;
    if ($chown and $chown =~ m/:/) {
        my @chownParts = split(':', $chown); # replaced down below
        $owner = shift @chownParts;
        $group = shift @chownParts;
    }

    my $search_and_replace_args;
    while ( my ( $search, $replace ) = each %{$translations} ) {
        if ($search =~ m/\?/) {
            $search_and_replace_args .= " -pe 's|${search}|${replace}|g;'";
        } else {
            $search_and_replace_args .= " -pe 's?${search}?${replace}?g;'";
        }
    }

    # the function below is similar to Admin.pm renderTemplate can we mod that instead?
    # p $source_to_target_paths;
    while ( my ( $source, $destination ) = each %{$source_to_target_paths} ) {
        # p $search_and_replace_args;
        if (! -f $source ) {
            &printMsg('ERROR', "file: $source does not exist or is unreadable");
            return 1;
        }

        if (( -f $source ) and (! $chown or ! $mode)) {
            my $stat = get_file_stat($source);
            unless($mode) {
                $mode = $stat->{mode};
            }
            unless($chown) {
                $owner = "$stat->{user}";
                $group = "$stat->{group}";
            }
        }

        #NOTE: due to descrepencies with install syntax on freebsd and ubuntu/rhel OS's lets just use cp, chmod, chown
        my $install = "$SUDO cp $source $destination && $SUDO chown ${owner}:${group} $destination && $SUDO chmod $mode $destination";
        printMsg( "DEBUG", "Installing $destination via: $install");
        if (system($install) == 0) {
            &printMsg('OK', "$install");
        } else {
            &printMsg('FAIL', "$install");
            return 1;
        }
        my $cmd = "$SUDO perl -i ${search_and_replace_args} $destination";
        printMsg( "DEBUG", "rendering Template search & replace");
        printMsg( "DEBUG", "Running in-place search & replace: $cmd");
        eval {
            system($cmd ) == 0 or die "System call failed: $cmd $!";
        };
        if ($@) {
            &printMsg('ERROR', "Failed running $cmd Err:$@");
            return 1;
        }
    }
    return 0;
}

# generic function for uninstalling our threatstop config for an APP (syslog-ng, rc.local etc..)
sub uninstall_app_config {
    my ($OPTS, $app_type) = @_;

    my $config = $OPTS->{"installed_${app_type}_config"} || '';
    if ($config and -f $config) {
        my $config_bak = "${config}.bak";
        if (-f $config_bak) {
            &printMsg('DEBUG', "removing $config_bak");
            eval { qx($SUDO rm $config_bak) };
            if ($@) {
                &printMsg('DEBUG', "Return status: $@");
            }
            delete $OPTS->{"installed_${app_type}_config"};
            write_conf_keys($OPTS);
        }
    } else {
        &printMsg('DEBUG', "no installed $app_type configuration found" . $config);
        return 0;
    }
    my $base_confd = (exists $OPTS->{"${app_type}_confd"} and $OPTS->{"${app_type}_confd"}) ? basename $OPTS->{"${app_type}_confd"} : '';
    if ($base_confd) {
        # check if we can simply remove an installed config
        &printMsg('DEBUG', "checking if $base_confd is in config path: $config.");
        if (index($config, $base_confd)) {
            #if if matches we just delete the config.
            &printMsg('DEBUG', "$base_confd is in config path: $config.");
            my $exit;
            # eval { qx($SUDO rm "$config") or die("Could not delete $config"); $exit = $?; };
            eval { system("$SUDO rm \"$config\" 2> /dev/null"); $exit = $?; };
            if ($exit != 0) {
                &printMsg('FAIL', 'Delete confd config: ' . $@ . ' exit status:' . $exit);
                sleep 2; # let pending IO settle
                (-f $config) ? return 1 : return 0; # return error if file exists still
            } else {
                &printMsg('OK', "$config removed");
                return 0;
            }
        }
    }

    # else we have to unmerge a conf file
    if (exists $OPTS->{"installed_${app_type}_config"} and exists $OPTS->{"installed_${app_type}_merge"}) {
        my $merge_file = $OPTS->{"installed_${app_type}_merge"};
        my $config_basename = basename $config;

        # we have to remove delta configuration from file.
        # copy config to tmp location
        eval { qx($SUDO cp -p $config ${merge_file}.uninstall) or die("Could not install merged $config"); };
        if ($@) {&printMsg('ERROR', $@); return 1; }
        # copy tmp to merge.tmp
        eval { qx($SUDO diff $config "$OPTS->{"installed_${app_type}_merge"}"|sed -e '/^[+@]/d'|cut -c2- > ${merge_file}.uninstall) or die('failed getting delta of config') };
        if ($@) {&printMsg('ERROR', $@); return 1; }
        # remove our config section from merge.tmp (to preserve original perms)
        # swap back.
        eval { qx($SUDO cp -p ${merge_file}.uninstall $config) or die("Could not install merged $config"); };
        if ($@) {&printMsg('ERROR', $@); return 1; }
    }
    elsif (exists $OPTS->{"installed_${app_type}_config"} and ! exists $OPTS->{"installed_${app_type}_merge"}) {
        # we have a conf.d install.. easily just remove our config.
        &printMsg('DEBUG', "$base_confd is in config path: $config.");
        my $exit;
        # eval { qx($SUDO rm "$config") or die("Could not delete $config"); $exit = $?; };
        eval { system("$SUDO rm \"$config\" 2> /dev/null"); $exit = $?; };
        if ($exit != 0) {
            &printMsg('FAIL', 'Delete confd config[2]: ' . $@ . ' exit status:' . $exit);
            sleep 2; # let pending IO settle
            (-f $config) ? return 1 : return 0; # return error if file exists still
        } else {
            &printMsg('OK', "$config removed");
            return 0;
        }
    } else {
        my $original_backup = get_backup_path($config) . '.orig';
        if (-f $original_backup) {
            &printMsg('DEBUG', "Backup config found: $original_backup");
            if (confirm_prompt('Would you like to restore the logging configuration from before installing ThreatSTOP?') ) {
                restore_file($config);
            }
        }
        # unknown state.
        &printMsg('DEBUG', "Logging does not seem to be installed");
    }

    my $restart_cmd = (exists $OPTS->{"${app_type}_restart_cmd"} and $OPTS->{"${app_type}_restart_cmd"}) ? $OPTS->{"${app_type}_restart_cmd"} : '';
    if ($restart_cmd) {
        eval { qx($SUDO $restart_cmd) or die('failed restarting ${app_type}') };
        if ($@) {
            &printMsg('DEBUG', "Logger uninstall completed & restarted");
        } else {
            &printMsg('WARN', "Problem detected restarting ${app_type}");
        }
    }
}

sub is_ctp_install {
    my ($OPTS) = @_;
    # auto config check
    if (exists $OPTS->{'auto_key'} and $OPTS->{'auto_key'} and ( (exists $OPTS->{'type'} and lc $OPTS->{'type'} eq 'auto' ) or (ThreatStop::LocalDevice::is_already_configured($OPTS)) )) {
        &printMsg('DEBUG', 'can_install: Detected auto config');
        return 1;
    }
    elsif (-f "$OPTS->{'src'}/.installed") {
        chomp(my $install_method = `cat $OPTS->{'src'}/.installed`);
        if ($install_method eq 'auto') {
            &printMsg('DEBUG', "reading .installed file flag Detected auto config");
            return 1;
        }
    }
    return 0;
}

sub get_file_mtime {
    my $file = shift;
    if (-f $file) {
        return (stat($file))[9];
    }
    return 0;
}

=head3 generate_checksum

take in string of all values concatinated together pipe to shasum -a 1|awk '{print $1}'

args:
    string - concatinated string of all values (CTP or Telemetry)
returns:
    sha1 hash

=cut

sub generate_sha1_checksum {
    my ($OPTS, $str) = @_;
    unless($str) {
        &printMsg('DEBUG', "Nothing passed to generate_checksum");
        return;
    }
    my $checksum_cmd = "echo -n '$str' | $OPTS->{shasum} -a 1 | cut -d' ' -f1";
    # &printMsg('DEBUG', "checksum cmd: $checksum_cmd");
    chomp( my $checksum = `$checksum_cmd` );
    &printMsg('INFO', "Generated checksum: $checksum");
    return $checksum;
}

=head3 set_f_telemetry

Simple routine to touch a hidden file to update it's mtime.
It is how we keep track of some telemetry items: (last_policy_update, config_change, last_log_upload_attempt, etc...)
all markers go in /opt/threatstop and are prepended with a dot (making them hidden)
    args:
        string - marker name
    returns:
            0 - success
            1 - fail
=cut
sub set_f_telemetry {
    my $marker = shift;
    my $value = shift;
    if ($value) {
        qx(echo "$value" > $TSDIR/.$marker);
    } else {
        qx(touch $TSDIR/.$marker);
    }
    return $?;
}

sub clear_f_telemetry {
    my $marker = shift;
    my $fpath = "$TSDIR/.$marker";
    return 0 unless (-f $fpath);
    qx(rm $TSDIR/.$marker 2> /dev/null);
    return $?;
}

sub get_f_telemetry {
    my $marker = shift;
    my $fpath = "$TSDIR/.$marker";
    return get_file_mtime($fpath) || 0;
}

sub get_f_telemetry_value {
    my $marker = shift;
    my $fpath = "$TSDIR/.$marker";
    if (-f $fpath and -s $fpath)  {
        my $value = `cat $fpath`;
        chomp($value);
        return $value if $value;
    }
    return '';
}
# get seconds since now
sub get_f_telemetry_elapsed {
    my $marker = shift;
    my $now = time;
    my $marker_mtime = get_f_telemetry($marker);
    return ($now - $marker_mtime);
}

=head3 seteuid

 set effective uid using $> variable.
 this is here mainly to make it more clear what we are doing instead of
 just setting a magic perl variable in the code.
=cut
sub seteuid {
    $> = shift;
    return $!;
}

###############################################################
##################### MODULE END ##############################
###############################################################

1;
