#$Id$
#Package Maintainer : support@threatstop.com
#########################################################################
# Copyright(c) 2006-2019 ThreatSTOP, Inc. All Rights Reserved
#
# NOTICE: All information contained herein is, and remains the property
# of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to ThreatSTOP, Inc.
# and its suppliers and certain aspects thereof are protected by United
# States Patent No. 8,533,822 and United States Patent No. 8,869,237,
# United States Patent No. 62539504 (Pending),
# and are protected by trade secret or copyright law.
#
# US Government Entities: The ThreatSTOP service, software and
# documentation, as applicable, are "commercial computer software" and
# "commercial computer software documentation" developed exclusively at
# private expense by Threatstop, Inc. ("Threatstop"). Pursuant to FAR
# 12.212 or DFARS 227 7202 and their successors, as applicable, use,
# reproduction and disclosure of the software and documentation is
# governed by the terms of Threatstop's commercial agreement.
########################################################################

# Wrapper for managing stunnel
# Deal with concurrency from multiple devices by tracking users of stunnel

package ThreatStop::STunnel;
use strict;
use warnings;
use 5.010;    #enable perl 5.10 features.

use ThreatStop qw($SUDO $TSDIR install_app_config printMsg);
use Data::Dumper;

# Standard Module defines
our $VERSION = '2.0';
our @ISA     = qw(Exporter);
our @EXPORT =
    qw(stunnel_required stunnel_completed );
our @EXPORT_OK = qw();


# Variations for starting/stopping services
# use constant {
#     UNSUPPORTED_OS => 0,
#     SYSTEMD_UBUNTU => 1,
#     SYSTEMD_REDHAT => 2,
#     SERVICE => 3,
#     VYATTA_BASED => 4,
# };

# DELETE ME
# my $dig=`sudo which dig|tr -d '\n'`;
# sub update_dig_location {
#     my $new_dig = shift;
#     if (-f $new_dig and $dig ne $new_dig) {
#         print "updating dig location: $new_dig\n";
#         $dig = $new_dig;
#     }
# }

my $STATUS_DIR="$TSDIR/tmp/stunnel";
if (! -d $STATUS_DIR) {
    system("mkdir -p $STATUS_DIR");
}

# my $STUNNEL_PID = '/var/run/stunnel.pid';
my $STUNNEL_PID = $TSDIR . '/stunnel.pid';
my $STUNNEL_LOG = $TSDIR . '/stunnel.log';
my $STUNNEL_CONF = $TSDIR . '/threatstop_stunnel.conf';
unless ( -f $STUNNEL_CONF ) {
    $STUNNEL_CONF = '/etc/stunnel/threatstop_stunnel.conf';
}


sub stunnel_required {
    my ($OPTS) = @_;

    # first check to see if we've installed our configuration file.
    my $stunnel_template = "$TSDIR/templates/stunnel/threatstop_stunnel.conf";
    my $search_and_replace = {
        __CTP_SERVER_PORT__ => $OPTS->{webauto_server_port},
        __STUNNEL_SERVER__ => $OPTS->{webauto_upstream_server},
        __TS_DIR__ => $TSDIR
    };
    unless (-f $STUNNEL_CONF) {
        &printMsg('DEBUG', "Installing stunnel configuration");
        return 1 if (install_app_config($OPTS, 'tunneler', $stunnel_template, $search_and_replace, 'threatstop:threatstop') != 0 );
    }

    my $PID_FILE = $STATUS_DIR . $$;

    if (! -e $STATUS_DIR) {
        &printMsg('DEBUG', "Creating stunnel status dir");
        mkdir($STATUS_DIR);
    }

    local $/ = undef;
    if (system("$SUDO echo 'active' > $PID_FILE") == 0 ) {
        &printMsg('OK', "setting stunnel pid 'active' $PID_FILE");
    } else {
        &printMsg('FAIL', "setting stunnel pid 'active' $PID_FILE");
    }
    # open FILE, ">$PID_FILE";
    # binmode FILE;
    # print FILE "active\n";
    # close FILE;

    # Always start it; even if it is already started
    &printMsg('DEBUG', "Attemping to start stunnel");
    if (_start_stunnel($OPTS) == 0 ) {
        &printMsg('UP', "Build Encrypted Channel");
        return 0;
    } else {
        &printMsg('DOWN', "Build Encrypted Channel");
        return 1;
    }
}

sub stunnel_completed {
    my ($OPTS) = @_;
    # Remove our own PID file
    my $PID_FILE = $STATUS_DIR . $$;
    unlink($PID_FILE);

    # Remove anything older than 10 minutes, in case something crashed
    # No CTP command should take than long
    my $can_stop = 1;
    opendir (my $dh, $STATUS_DIR);
    my @remaining_files = readdir $dh;
    foreach my $file (@remaining_files) {
        next if $file eq "." or $file eq "..";
        if (time - (stat $STATUS_DIR.$file)[9] > 10 * 60) {
            unlink($STATUS_DIR.$file);
        } else {
            $can_stop = 0;
        }
    }
    if ($can_stop) {
        &printMsg('DOWN', "Teardown Encrypted Channel");
        if (_stop_stunnel() == 0 ) {
            &printMsg('OK', "Teardown Encrypted Channel");
        } else {
            &printMsg('FAIL', "Teardown Encrypted Channel");
        }
    }
}


sub _start_stunnel {
    my ($OPTS) = @_;

    my $stunnel = $OPTS->{stunnel4} ? $OPTS->{stunnel4} : $OPTS->{stunnel};
    my $should_start = 1;


    if (-e $STUNNEL_PID) {
        my $stunpid;
        open my $fh, '<', $STUNNEL_PID;
        $stunpid = do { local $/; <$fh> };
        close $fh;
        chomp $stunpid;
        system("$SUDO kill -0 $stunpid");
        # Got a PID file, got a running process, don't start a new one
        if ($? eq 0) { $should_start = 0; }
    }
    if ($should_start) {
        &printMsg('START', "Encrypted channel");
        my $start_cmd = "$SUDO $stunnel $STUNNEL_CONF > $STUNNEL_LOG 2>&1";
        if ( system($start_cmd) == 0 ) {
            &printMsg('UP', "Encrypted channel");
            &printMsg('DEBUG', "$start_cmd");
        } else {
            &printMsg('FAIL', "Encrypted channel");
            &printMsg('DEBUG', "$start_cmd");
        }
    }

    # Check that stunnel has settled
    my $stunnel_ok = 0;
    my $attempts = 0;
    my $cmd = "$OPTS->{dig} +time=1 +tcp -p 5353 -t TXT \@127.0.0.1 ping.ctp.threatstop.local +short";
    # print "DEBUG CMD: $cmd\n";
    while (! $stunnel_ok and $attempts < 10) {
        &printMsg('DEBUG', "Running: $cmd");
        my $cmd_out = `$cmd`;
        if ($cmd_out =~ /PONG/i ) {
            &printMsg('DEBUG', "tunnel up and running");
            return 0;
        }
        sleep(1);
        $attempts++;
    }
    &printMsg('ERROR', "Failed to connect to ThreatSTOP CTP DNS Service\n");
    # CTP requests will fail
    return 1;
}

sub _stop_stunnel {

    if (-e $STUNNEL_PID) {
        if (system("$SUDO /bin/kill `cat $STUNNEL_PID`") == 0 ) {
            return 0;
        }
        return 1;
    }

}

1;
